@@ -32,10 +32,12 @@
 #include <arpa/inet.h>
 #include <errno.h>
 #include <stdlib.h>
-#include <glib/gi18n-lib.h>
 #include <string.h>
+#include <sys/stat.h>
+#include <fcntl.h>
+
+#include <glib/gi18n-lib.h>
 #include <gtk/gtk.h>
-#include <gnome-keyring.h>
 #include <gnome-keyring-memory.h>
 
 #define NM_VPN_API_SUBJECT_TO_CHANGE
@@ -46,7 +48,6 @@
 #include <nm-setting-ip4-config.h>
 
 #include "src/nm-vpnc-service.h"
-#include "common-gnome/keyring-helpers.h"
 #include "pcf-file.h"
 #include "nm-vpnc.h"
 
@@ -87,6 +88,7 @@ typedef struct {
 	GtkWidget *widget;
 	GtkSizeGroup *group;
 	gint orig_dpd_timeout;
+	GtkWidget *advanced_dialog;
 } VpncPluginUiWidgetPrivate;
 
 
@@ -165,80 +167,54 @@ stuff_changed_cb (GtkWidget *widget, gpointer user_data)
 	g_signal_emit_by_name (VPNC_PLUGIN_UI_WIDGET (user_data), "changed");
 }
 
-static gboolean
-fill_vpn_passwords (VpncPluginUiWidget *self, NMConnection *connection)
+static void
+hybrid_toggled_cb (GtkWidget *widget, gpointer user_data)
+{
+	VpncPluginUiWidgetPrivate *priv = VPNC_PLUGIN_UI_WIDGET_GET_PRIVATE (user_data);
+	gboolean enabled = FALSE;
+	GtkWidget *cafile_label, *ca_file_chooser;
+
+	cafile_label = GTK_WIDGET (gtk_builder_get_object (priv->builder, "cafile_label"));
+	g_return_if_fail (cafile_label);
+	ca_file_chooser = GTK_WIDGET (gtk_builder_get_object (priv->builder, "ca_file_chooser"));
+	g_return_if_fail (ca_file_chooser);
+
+	enabled = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));
+
+	gtk_widget_set_sensitive (cafile_label, enabled);
+	gtk_widget_set_sensitive (ca_file_chooser, enabled);
+
+	stuff_changed_cb (widget, user_data);
+}
+
+static void
+setup_password_widget (VpncPluginUiWidget *self,
+                       const char *entry_name,
+                       NMSettingVPN *s_vpn,
+                       const char *secret_name,
+                       gboolean new_connection)
 {
 	VpncPluginUiWidgetPrivate *priv = VPNC_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
+	NMSettingSecretFlags secret_flags = NM_SETTING_SECRET_FLAG_NONE;
 	GtkWidget *widget;
-	gboolean success = FALSE;
-	char *password = NULL;
-	char *group_password = NULL;
+	const char *value;
 
-	/* Grab secrets from the keyring */
-	if (connection) {
-		NMSettingConnection *s_con;
-		NMSettingVPN *s_vpn;
-		const char *tmp;
+	if (new_connection)
+		secret_flags = NM_SETTING_SECRET_FLAG_AGENT_OWNED;
 
-		s_vpn = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);
+	widget = (GtkWidget *) gtk_builder_get_object (priv->builder, entry_name);
+	g_assert (widget);
+	gtk_size_group_add_widget (priv->group, widget);
 
-		if (nm_connection_get_scope (connection) == NM_CONNECTION_SCOPE_SYSTEM) {
-			if (s_vpn) {
-				tmp = nm_setting_vpn_get_secret (s_vpn, NM_VPNC_KEY_XAUTH_PASSWORD);
-				if (tmp)
-					password = gnome_keyring_memory_strdup (tmp);
-
-				tmp = nm_setting_vpn_get_secret (s_vpn, NM_VPNC_KEY_SECRET);
-				if (tmp)
-					group_password = gnome_keyring_memory_strdup (tmp);
-			}
-		} else {
-			s_con = NM_SETTING_CONNECTION (nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION));
-
-			/* Lookup passwords in the keyring, and if they weren't there, try
-			 * the connection itself, which is where they'd be right after import.
-			 */
-			keyring_helpers_get_one_secret (nm_setting_connection_get_uuid (s_con),
-			                                VPNC_USER_PASSWORD,
-			                                &password,
-			                                NULL);
-			if (!password)
-				password = gnome_keyring_memory_strdup (nm_setting_vpn_get_secret (s_vpn, NM_VPNC_KEY_XAUTH_PASSWORD));
-
-			keyring_helpers_get_one_secret (nm_setting_connection_get_uuid (s_con),
-			                                VPNC_GROUP_PASSWORD,
-			                                &group_password,
-			                                NULL);
-			if (!group_password)
-				group_password = gnome_keyring_memory_strdup (nm_setting_vpn_get_secret (s_vpn, NM_VPNC_KEY_SECRET));
-		}
+	if (s_vpn) {
+		value = nm_setting_vpn_get_secret (s_vpn, secret_name);
+		gtk_entry_set_text (GTK_ENTRY (widget), value ? value : "");
+		nm_setting_get_secret_flags (NM_SETTING (s_vpn), secret_name, &secret_flags, NULL);
 	}
+	secret_flags &= ~(NM_SETTING_SECRET_FLAG_NOT_SAVED | NM_SETTING_SECRET_FLAG_NOT_REQUIRED);
+	g_object_set_data (G_OBJECT (widget), "flags", GUINT_TO_POINTER (secret_flags));
 
-	/* User password */
-	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "user_password_entry"));
-	if (!widget)
-		goto out;
-	if (password)
-		gtk_entry_set_text (GTK_ENTRY (widget), password);
-	gtk_size_group_add_widget (priv->group, GTK_WIDGET (widget));
-	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);
-
-	/* Group password */
-	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "group_password_entry"));
-	if (!widget)
-		goto out;
-	if (group_password)
-		gtk_entry_set_text (GTK_ENTRY (widget), group_password);
-	gtk_size_group_add_widget (priv->group, GTK_WIDGET (widget));
-	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);
-
-	success = TRUE;
-
-out:
-	gnome_keyring_memory_free (password);
-	gnome_keyring_memory_free (group_password);
-
-	return success;
+	g_signal_connect (widget, "changed", G_CALLBACK (stuff_changed_cb), self);
 }
 
 static void
@@ -265,7 +241,6 @@ pw_type_changed_helper (VpncPluginUiWidget *self, GtkWidget *combo)
 	VpncPluginUiWidgetPrivate *priv = VPNC_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
 	const char *entry = NULL;
 	GtkWidget *widget;
-	GtkTreeModel *model;
 
 	/* If the user chose "Not required", desensitize and clear the correct
 	 * password entry.
@@ -284,7 +259,6 @@ pw_type_changed_helper (VpncPluginUiWidget *self, GtkWidget *combo)
 	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, entry));
 	g_assert (widget);
 
-	model = gtk_combo_box_get_model (GTK_COMBO_BOX (combo));
 	switch (gtk_combo_box_get_active (GTK_COMBO_BOX (combo))) {
 	case PW_TYPE_ASK:
 	case PW_TYPE_UNUSED:
@@ -306,11 +280,27 @@ pw_type_combo_changed_cb (GtkWidget *combo, gpointer user_data)
 	stuff_changed_cb (combo, self);
 }
 
+static const char *
+secret_flags_to_pw_type (NMSettingVPN *s_vpn, const char *key)
+{
+	NMSettingSecretFlags flags = NM_SETTING_SECRET_FLAG_NONE;
+
+	if (nm_setting_get_secret_flags (NM_SETTING (s_vpn), key, &flags, NULL)) {
+		if (flags & NM_SETTING_SECRET_FLAG_NOT_REQUIRED)
+			return NM_VPNC_PW_TYPE_UNUSED;
+		if (flags & NM_SETTING_SECRET_FLAG_NOT_SAVED)
+			return NM_VPNC_PW_TYPE_ASK;
+		return NM_VPNC_PW_TYPE_SAVE;
+	}
+	return NULL;
+}
+
 static void
 init_one_pw_combo (VpncPluginUiWidget *self,
                    NMSettingVPN *s_vpn,
                    const char *combo_name,
-                   const char *key,
+                   const char *secret_key,
+                   const char *type_key,
                    const char *entry_name)
 {
 	VpncPluginUiWidgetPrivate *priv = VPNC_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
@@ -325,17 +315,17 @@ init_one_pw_combo (VpncPluginUiWidget *self,
 	 * the VPN settings, default to saving it.  Otherwise, always ask for it.
 	 */
 	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, entry_name));
-	if (widget) {
-		const char *tmp;
-
-		tmp = gtk_entry_get_text (GTK_ENTRY (widget));
-		if (tmp && strlen (tmp))
-			default_idx = 0;
-	}
+	g_assert (widget);
+	value = gtk_entry_get_text (GTK_ENTRY (widget));
+	if (value && strlen (value))
+		default_idx = 0;
 
 	store = gtk_list_store_new (1, G_TYPE_STRING);
-	if (s_vpn)
-		value = nm_setting_vpn_get_data_item (s_vpn, key);
+	if (s_vpn) {
+		value = secret_flags_to_pw_type (s_vpn, secret_key);
+		if (!value)
+			value = nm_setting_vpn_get_data_item (s_vpn, type_key);
+	}
 
 	gtk_list_store_append (store, &iter);
 	gtk_list_store_set (store, &iter, 0, _("Saved"), -1);
@@ -368,11 +358,95 @@ init_one_pw_combo (VpncPluginUiWidget *self,
 	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (pw_type_combo_changed_cb), self);
 }
 
+static void
+toggle_advanced_dialog_cb (GtkWidget *button, gpointer user_data)
+{
+	VpncPluginUiWidgetPrivate *priv = VPNC_PLUGIN_UI_WIDGET_GET_PRIVATE (user_data);
+	GtkWidget *toplevel;
+
+	if (gtk_widget_get_visible (priv->advanced_dialog))
+		gtk_widget_hide (priv->advanced_dialog);
+	else {
+		toplevel = gtk_widget_get_toplevel (priv->widget);
+		if (gtk_widget_is_toplevel (toplevel))
+			gtk_window_set_transient_for (GTK_WINDOW (priv->advanced_dialog), GTK_WINDOW (toplevel));
+		gtk_widget_show_all (priv->advanced_dialog);
+	}
+}
+
+static const char *
+find_tag (const char *tag, const char *buf, gsize len)
+{
+	gsize i, taglen;
+
+	taglen = strlen (tag);
+	if (len < taglen)
+		return NULL;
+
+	for (i = 0; i < len - taglen + 1; i++) {
+		if (memcmp (buf + i, tag, taglen) == 0)
+			return buf + i;
+	}
+	return NULL;
+}
+
+static const char *pem_cert_begin = "-----BEGIN CERTIFICATE-----";
+
 static gboolean
-init_plugin_ui (VpncPluginUiWidget *self, NMConnection *connection, GError **error)
+cert_filter (const GtkFileFilterInfo *filter_info, gpointer data)
+{
+	char *contents = NULL, *p, *ext;
+	gsize bytes_read = 0;
+	gboolean show = FALSE;
+	struct stat statbuf;
+
+	if (!filter_info->filename)
+		return FALSE;
+
+	p = strrchr (filter_info->filename, '.');
+	if (!p)
+		return FALSE;
+
+	ext = g_ascii_strdown (p, -1);
+	if (!ext)
+		return FALSE;
+
+	if (strcmp (ext, ".pem") && strcmp (ext, ".crt") && strcmp (ext, ".cer")) {
+		g_free (ext);
+		return FALSE;
+	}
+	g_free (ext);
+
+	/* Ignore files that are really large */
+	if (!stat (filter_info->filename, &statbuf)) {
+		if (statbuf.st_size > 500000)
+			return FALSE;
+	}
+
+	if (!g_file_get_contents (filter_info->filename, &contents, &bytes_read, NULL))
+		return FALSE;
+
+	if (bytes_read < 400)  /* needs to be lower? */
+		goto out;
+
+	if (find_tag (pem_cert_begin, (const char *) contents, bytes_read)) {
+		show = TRUE;
+		goto out;
+	}
+
+out:
+	g_free (contents);
+	return show;
+}
+
+static gboolean
+init_plugin_ui (VpncPluginUiWidget *self,
+                NMConnection *connection,
+                gboolean new_connection,
+                GError **error)
 {
 	VpncPluginUiWidgetPrivate *priv = VPNC_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
-	NMSettingVPN *s_vpn;
+	NMSettingVPN *s_vpn = NULL;
 	GtkWidget *widget;
 	GtkListStore *store;
 	GtkTreeIter iter;
@@ -380,8 +454,11 @@ init_plugin_ui (VpncPluginUiWidget *self, NMConnection *connection, GError **err
 	int active = -1;
 	const char *natt_mode = NULL;
 	const char *ike_dh_group = NULL;
+	gboolean enabled = FALSE;
+	GtkFileFilter *filter;
 
-	s_vpn = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);
+	if (connection)
+		s_vpn = nm_connection_get_setting_vpn (connection);
 
 	priv->group = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);
 
@@ -438,12 +515,29 @@ init_plugin_ui (VpncPluginUiWidget *self, NMConnection *connection, GError **err
 	/* Fill the VPN passwords *before* initializing the PW type combos, since
 	 * knowing if there are passwords when initializing the combos is helpful.
 	 */
-	fill_vpn_passwords (self, connection);
+	setup_password_widget (self,
+	                       "user_password_entry",
+	                       s_vpn,
+	                       NM_VPNC_KEY_XAUTH_PASSWORD,
+	                       new_connection);
+	setup_password_widget (self,
+	                       "group_password_entry",
+	                       s_vpn,
+	                       NM_VPNC_KEY_SECRET,
+	                       new_connection);
 
-	init_one_pw_combo (self, s_vpn, "user_pass_type_combo",
-	                   NM_VPNC_KEY_XAUTH_PASSWORD_TYPE, "user_password_entry");
-	init_one_pw_combo (self, s_vpn, "group_pass_type_combo",
-	                   NM_VPNC_KEY_SECRET_TYPE, "group_password_entry");
+	init_one_pw_combo (self,
+	                   s_vpn,
+	                   "user_pass_type_combo",
+	                   NM_VPNC_KEY_XAUTH_PASSWORD,
+	                   NM_VPNC_KEY_XAUTH_PASSWORD_TYPE,
+	                   "user_password_entry");
+	init_one_pw_combo (self,
+	                   s_vpn,
+	                   "group_pass_type_combo",
+	                   NM_VPNC_KEY_SECRET,
+	                   NM_VPNC_KEY_SECRET_TYPE,
+	                   "group_password_entry");
 
 	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "user_entry"));
 	g_return_val_if_fail (widget != NULL, FALSE);
@@ -540,7 +634,6 @@ init_plugin_ui (VpncPluginUiWidget *self, NMConnection *connection, GError **err
 	gtk_combo_box_set_active (GTK_COMBO_BOX (widget), active < 0 ? 1 : active);
 	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);
 
-
 	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "disable_dpd_checkbutton"));
 	g_return_val_if_fail (widget != NULL, FALSE);
 	if (s_vpn) {
@@ -565,6 +658,60 @@ init_plugin_ui (VpncPluginUiWidget *self, NMConnection *connection, GError **err
 	                  (GCallback) show_toggled_cb,
 	                  self);
 
+	/* hybrid auth */
+
+	enabled = FALSE;
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "hybrid_checkbutton"));
+	g_return_val_if_fail (widget != NULL, FALSE);
+	if (s_vpn) {
+		value = nm_setting_vpn_get_data_item (s_vpn, NM_VPNC_KEY_AUTHMODE);
+		if (value && !strcmp("hybrid", value)) {
+			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
+			enabled = TRUE;
+		}
+	}
+	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (hybrid_toggled_cb), self);
+
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "ca_file_chooser"));
+	g_return_val_if_fail (widget != NULL, FALSE);
+	gtk_size_group_add_widget (priv->group, widget);
+	gtk_widget_set_sensitive (widget, enabled);
+	gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (widget), TRUE);
+	gtk_file_chooser_button_set_title (GTK_FILE_CHOOSER_BUTTON (widget),
+	                                   _("Choose a Certificate Authority (CA) certificate..."));
+
+	filter = gtk_file_filter_new ();
+	gtk_file_filter_add_custom (filter, GTK_FILE_FILTER_FILENAME, cert_filter, NULL, NULL);
+	gtk_file_filter_set_name (filter, _("PEM certificates (*.pem, *.crt, *.cer)"));
+	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (widget), filter);
+
+	if (s_vpn) {
+		value = nm_setting_vpn_get_data_item (s_vpn, NM_VPNC_KEY_CA_FILE);
+		if (value && strlen (value))
+			gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (widget), value);
+	}
+	g_signal_connect (G_OBJECT (widget), "file-set", G_CALLBACK (stuff_changed_cb), self);
+
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "cafile_label"));
+	g_return_val_if_fail (widget != NULL, FALSE);
+	gtk_widget_set_sensitive (widget, enabled);
+
+	/* advanced dialog */
+
+	priv->advanced_dialog = GTK_WIDGET (gtk_builder_get_object (priv->builder, "vpnc-advanced-dialog"));
+	g_return_val_if_fail (priv->advanced_dialog != NULL, FALSE);
+
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "advanced_button"));
+	g_return_val_if_fail (widget != NULL, FALSE);
+	g_signal_connect (G_OBJECT (widget), "clicked",
+	                  (GCallback) toggle_advanced_dialog_cb,
+	                  self);
+
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "apply_button"));
+	g_return_val_if_fail (widget != NULL, FALSE);
+	g_signal_connect (G_OBJECT (widget), "clicked",
+	                  (GCallback) toggle_advanced_dialog_cb,
+	                  self);
 	return TRUE;
 }
 
@@ -577,32 +724,46 @@ get_widget (NMVpnPluginUiWidgetInterface *iface)
 	return G_OBJECT (priv->widget);
 }
 
-static guint32
-handle_one_pw_type (NMSettingVPN *s_vpn, GtkBuilder *builder, const char *name, const char *key)
+static void
+save_one_password (NMSettingVPN *s_vpn,
+                   GtkBuilder *builder,
+                   const char *entry_name,
+                   const char *combo_name,
+                   const char *secret_key,
+                   const char *type_key)
 {
-	GtkWidget *widget;
-	GtkTreeModel *model;
-	guint32 pw_type;
+	NMSettingSecretFlags flags = NM_SETTING_SECRET_FLAG_NONE;
+	const char *data_val = NULL, *password;
+	GtkWidget *entry;
+	GtkWidget *combo;
 
-	widget = GTK_WIDGET (gtk_builder_get_object (builder, name));
-	model = gtk_combo_box_get_model (GTK_COMBO_BOX (widget));
+	/* Grab original password flags */
+	entry = GTK_WIDGET (gtk_builder_get_object (builder, entry_name));
+	flags = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (entry), "flags"));
 
-	pw_type = gtk_combo_box_get_active (GTK_COMBO_BOX (widget));
-	switch (pw_type) {
+	/* And set new ones based on the type combo */
+	combo = GTK_WIDGET (gtk_builder_get_object (builder, combo_name));
+	switch (gtk_combo_box_get_active (GTK_COMBO_BOX (combo))) {
 	case PW_TYPE_SAVE:
-		nm_setting_vpn_add_data_item (s_vpn, key, NM_VPNC_PW_TYPE_SAVE);
+		password = gtk_entry_get_text (GTK_ENTRY (entry));
+		if (password && strlen (password))
+			nm_setting_vpn_add_secret (s_vpn, secret_key, password);
+		data_val = NM_VPNC_PW_TYPE_SAVE;
 		break;
 	case PW_TYPE_UNUSED:
-		nm_setting_vpn_add_data_item (s_vpn, key, NM_VPNC_PW_TYPE_UNUSED);
+		data_val = NM_VPNC_PW_TYPE_UNUSED;
+		flags |= NM_SETTING_SECRET_FLAG_NOT_REQUIRED;
 		break;
 	case PW_TYPE_ASK:
 	default:
-		pw_type = PW_TYPE_ASK;
-		nm_setting_vpn_add_data_item (s_vpn, key, NM_VPNC_PW_TYPE_ASK);
+		data_val = NM_VPNC_PW_TYPE_ASK;
+		flags |= NM_SETTING_SECRET_FLAG_NOT_SAVED;
 		break;
 	}
 
-	return pw_type;
+	/* Set both new secret flags and old data item for backwards compat */
+	nm_setting_vpn_add_data_item (s_vpn, type_key, data_val);
+	nm_setting_set_secret_flags (NM_SETTING (s_vpn), secret_key, flags, NULL);
 }
 
 static gboolean
@@ -617,7 +778,6 @@ update_connection (NMVpnPluginUiWidgetInterface *iface,
 	char *str;
 	GtkTreeModel *model;
 	GtkTreeIter iter;
-	guint32 upw_type, gpw_type;
 
 	if (!check_validity (self, error))
 		return FALSE;
@@ -695,103 +855,53 @@ update_connection (NMVpnPluginUiWidgetInterface *iface,
 		}
 	}
 
-	upw_type = handle_one_pw_type (s_vpn, priv->builder, "user_pass_type_combo", NM_VPNC_KEY_XAUTH_PASSWORD_TYPE);
-	gpw_type = handle_one_pw_type (s_vpn, priv->builder, "group_pass_type_combo", NM_VPNC_KEY_SECRET_TYPE);
+	/* User password */
+	save_one_password (s_vpn,
+	                   priv->builder,
+	                   "user_password_entry",
+	                   "user_pass_type_combo",
+	                   NM_VPNC_KEY_XAUTH_PASSWORD,
+	                   NM_VPNC_KEY_XAUTH_PASSWORD_TYPE);
 
-	/* System secrets get stored in the connection, user secrets are saved
-	 * via the save_secrets() hook.
-	 */
-	if (nm_connection_get_scope (connection) == NM_CONNECTION_SCOPE_SYSTEM) {
-		/* User password */
-		widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "user_password_entry"));
-		str = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
-		if (str && strlen (str) && (upw_type != PW_TYPE_UNUSED))
-			nm_setting_vpn_add_secret (s_vpn, NM_VPNC_KEY_XAUTH_PASSWORD, str);
-
-		/* Group password */
-		widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "group_password_entry"));
-		str = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
-		if (str && strlen (str) && (gpw_type != PW_TYPE_UNUSED))
-			nm_setting_vpn_add_secret (s_vpn, NM_VPNC_KEY_SECRET, str);
-	}
+	/* Group password */
+	save_one_password (s_vpn,
+	                   priv->builder,
+	                   "group_password_entry",
+	                   "group_pass_type_combo",
+	                   NM_VPNC_KEY_SECRET,
+	                   NM_VPNC_KEY_SECRET_TYPE);
 
 	/* Local Port is not in GUI (yet?). So when present in the connection,
 	 * copy it from the old VPN setting to the new one to preserve it.
 	 */
-	s_vpn_orig = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);
+	s_vpn_orig = nm_connection_get_setting_vpn (connection);
 	if (s_vpn_orig) {
 		const char *local_port = nm_setting_vpn_get_data_item (s_vpn_orig, NM_VPNC_KEY_LOCAL_PORT);
 		if (local_port && strlen (local_port))
 			nm_setting_vpn_add_data_item (s_vpn, NM_VPNC_KEY_LOCAL_PORT, local_port);
 	}
 
+	/* hybrid auth */
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "hybrid_checkbutton"));
+	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
+		nm_setting_vpn_add_data_item (s_vpn, NM_VPNC_KEY_AUTHMODE, "hybrid");
+
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "ca_file_chooser"));
+	str = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (widget));
+	if (str && strlen (str))
+		nm_setting_vpn_add_data_item (s_vpn, NM_VPNC_KEY_CA_FILE, str);
+
 	nm_connection_add_setting (connection, NM_SETTING (s_vpn));
 	return TRUE;
 }
 
 static void
-save_one_password (GtkBuilder *builder,
-                   const char *keyring_tag,
-                   const char *uuid,
-                   const char *id,
-                   const char *entry,
-                   const char *combo,
-                   const char *desc)
+is_new_func (const char *key, const char *value, gpointer user_data)
 {
-	GnomeKeyringResult ret;
-	GtkWidget *widget;
-	const char *password;
-	GtkTreeModel *model;
-	gboolean saved = FALSE;
+	gboolean *is_new = user_data;
 
-	widget = GTK_WIDGET (gtk_builder_get_object (builder, combo));
-	g_assert (widget);
-	model = gtk_combo_box_get_model (GTK_COMBO_BOX (widget));
-	if (gtk_combo_box_get_active (GTK_COMBO_BOX (widget)) == PW_TYPE_SAVE) {
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, entry));
-		g_assert (widget);
-		password = gtk_entry_get_text (GTK_ENTRY (widget));
-		if (password && strlen (password)) {
-			ret = keyring_helpers_save_secret (uuid, id, NULL, keyring_tag, password);
-			if (ret == GNOME_KEYRING_RESULT_OK)
-				saved = TRUE;
-			else
-				g_warning ("%s: failed to save %s to keyring.", __func__, desc);
-		}
-	}
-
-	if (!saved)
-		keyring_helpers_delete_secret (uuid, keyring_tag);
-}
-
-static gboolean
-save_secrets (NMVpnPluginUiWidgetInterface *iface,
-              NMConnection *connection,
-              GError **error)
-{
-	VpncPluginUiWidget *self = VPNC_PLUGIN_UI_WIDGET (iface);
-	VpncPluginUiWidgetPrivate *priv = VPNC_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
-	NMSettingConnection *s_con;
-	const char *id, *uuid;
-
-	s_con = (NMSettingConnection *) nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION);
-	if (!s_con) {
-		g_set_error (error,
-		             VPNC_PLUGIN_UI_ERROR,
-		             VPNC_PLUGIN_UI_ERROR_INVALID_CONNECTION,
-		             "missing 'connection' setting");
-		return FALSE;
-	}
-
-	id = nm_setting_connection_get_id (s_con);
-	uuid = nm_setting_connection_get_uuid (s_con);
-
-	save_one_password (priv->builder, VPNC_USER_PASSWORD, uuid, id,
-	                   "user_password_entry", "user_pass_type_combo", "user password");
-	save_one_password (priv->builder, VPNC_GROUP_PASSWORD, uuid, id,
-	                   "group_password_entry", "group_pass_type_combo", "group password");
-
-	return TRUE;
+	/* If there are any VPN data items the connection isn't new */
+	*is_new = FALSE;
 }
 
 static NMVpnPluginUiWidgetInterface *
@@ -800,6 +910,8 @@ nm_vpn_plugin_ui_widget_interface_new (NMConnection *connection, GError **error)
 	NMVpnPluginUiWidgetInterface *object;
 	VpncPluginUiWidgetPrivate *priv;
 	char *ui_file;
+	NMSettingVPN *s_vpn;
+	gboolean is_new = TRUE;
 
 	if (error)
 		g_return_val_if_fail (*error == NULL, NULL);
@@ -815,6 +927,8 @@ nm_vpn_plugin_ui_widget_interface_new (NMConnection *connection, GError **error)
 	ui_file = g_strdup_printf ("%s/%s", UIDIR, "nm-vpnc-dialog.ui");
 	priv->builder = gtk_builder_new ();
 
+	gtk_builder_set_translation_domain (priv->builder, GETTEXT_PACKAGE);
+
 	if (!gtk_builder_add_from_file (priv->builder, ui_file, error)) {
 		g_warning ("Couldn't load builder file: %s",
 		           error && *error ? (*error)->message : "(unknown)");
@@ -827,8 +941,6 @@ nm_vpn_plugin_ui_widget_interface_new (NMConnection *connection, GError **error)
 	}
 	g_free (ui_file);
 
-	gtk_builder_set_translation_domain (priv->builder, GETTEXT_PACKAGE);
-
 	priv->widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "vpnc-vbox"));
 	if (!priv->widget) {
 		g_set_error (error, VPNC_PLUGIN_UI_ERROR, 0, "could not load UI widget");
@@ -837,7 +949,11 @@ nm_vpn_plugin_ui_widget_interface_new (NMConnection *connection, GError **error)
 	}
 	g_object_ref_sink (priv->widget);
 
-	if (!init_plugin_ui (VPNC_PLUGIN_UI_WIDGET (object), connection, error)) {
+	s_vpn = nm_connection_get_setting_vpn (connection);
+	if (s_vpn)
+		nm_setting_vpn_foreach_data_item (s_vpn, is_new_func, &is_new);
+
+	if (!init_plugin_ui (VPNC_PLUGIN_UI_WIDGET (object), connection, is_new, error)) {
 		g_object_unref (object);
 		return NULL;
 	}
@@ -857,6 +973,9 @@ dispose (GObject *object)
 	if (priv->widget)
 		g_object_unref (priv->widget);
 
+	if (priv->advanced_dialog)
+		gtk_widget_destroy (priv->advanced_dialog);
+
 	if (priv->builder)
 		g_object_unref (priv->builder);
 
@@ -884,7 +1003,6 @@ vpnc_plugin_ui_widget_interface_init (NMVpnPluginUiWidgetInterface *iface_class)
 	/* interface implementation */
 	iface_class->get_widget = get_widget;
 	iface_class->update_connection = update_connection;
-	iface_class->save_secrets = save_secrets;
 }
 
 static void
@@ -1097,20 +1215,37 @@ import (NMVpnPluginUiInterface *iface, const char *path, GError **error)
 	if (pcf_file_lookup_string (pcf, "main", "UserName", &buf))
 		nm_setting_vpn_add_data_item (s_vpn, NM_VPNC_KEY_XAUTH_USER, buf);
 
-	if (pcf_file_lookup_string (pcf, "main", "UserPassword", &buf))
+	if (pcf_file_lookup_string (pcf, "main", "UserPassword", &buf)) {
 		nm_setting_vpn_add_secret (s_vpn, NM_VPNC_KEY_XAUTH_PASSWORD, buf);
+		nm_setting_set_secret_flags (NM_SETTING (s_vpn),
+		                             NM_VPNC_KEY_XAUTH_PASSWORD,
+		                             NM_SETTING_SECRET_FLAG_AGENT_OWNED,
+		                             NULL);
+	}
 
 	if (pcf_file_lookup_bool (pcf, "main", "SaveUserPassword", &bool_value)) {
+		NMSettingSecretFlags flags = NM_SETTING_SECRET_FLAG_AGENT_OWNED;
+
 		if (bool_value) {
 			nm_setting_vpn_add_data_item (s_vpn,
 			                              NM_VPNC_KEY_XAUTH_PASSWORD_TYPE,
 			                              NM_VPNC_PW_TYPE_SAVE);
-		}
+		} else
+			flags |= NM_SETTING_SECRET_FLAG_NOT_SAVED;
+
+		nm_setting_set_secret_flags (NM_SETTING (s_vpn),
+			                         NM_VPNC_KEY_XAUTH_PASSWORD,
+			                         flags,
+			                         NULL);
 	}
 
-	if (pcf_file_lookup_string (pcf, "main", "GroupPwd", &buf))
+	if (pcf_file_lookup_string (pcf, "main", "GroupPwd", &buf)) {
 		nm_setting_vpn_add_secret (s_vpn, NM_VPNC_KEY_SECRET, buf);
-	else {
+		nm_setting_set_secret_flags (NM_SETTING (s_vpn),
+		                             NM_VPNC_KEY_SECRET,
+		                             NM_SETTING_SECRET_FLAG_AGENT_OWNED,
+		                             NULL);
+	} else {
 		/* Handle encrypted passwords */
 		if (pcf_file_lookup_string (pcf, "main", "enc_GroupPwd", &buf)) {
 			char *decrypted;
@@ -1120,10 +1255,34 @@ import (NMVpnPluginUiInterface *iface, const char *path, GError **error)
 				nm_setting_vpn_add_secret (s_vpn, NM_VPNC_KEY_SECRET, decrypted);
 				memset (decrypted, 0, strlen (decrypted));
 				g_free (decrypted);
+
+				nm_setting_set_secret_flags (NM_SETTING (s_vpn),
+				                             NM_VPNC_KEY_SECRET,
+				                             NM_SETTING_SECRET_FLAG_AGENT_OWNED,
+				                             NULL);
 			}
 		}
 	}
 
+	/* Group Password Flags */
+	if (pcf_file_lookup_bool (pcf, "main", "X-NM-SaveGroupPassword", &bool_value)) {
+		NMSettingSecretFlags flags = NM_SETTING_SECRET_FLAG_AGENT_OWNED;
+
+		if (bool_value) {
+			nm_setting_vpn_add_data_item (s_vpn,
+			                              NM_VPNC_KEY_SECRET_TYPE,
+			                              NM_VPNC_PW_TYPE_SAVE);
+		} else
+			flags |= NM_SETTING_SECRET_FLAG_NOT_SAVED;
+
+		nm_setting_set_secret_flags (NM_SETTING (s_vpn), NM_VPNC_KEY_SECRET, flags, NULL);
+	} else {
+		/* If the key isn't present, assume "saved" */
+		nm_setting_vpn_add_data_item (s_vpn,
+		                              NM_VPNC_KEY_SECRET_TYPE,
+		                              NM_VPNC_PW_TYPE_SAVE);
+	}
+
 	if (pcf_file_lookup_string (pcf, "main", "NTDomain", &buf))
 		nm_setting_vpn_add_data_item (s_vpn, NM_VPNC_KEY_DOMAIN, buf);
 
@@ -1249,13 +1408,16 @@ export (NMVpnPluginUiInterface *iface,
 	const char *domain = NULL;
 	const char *peertimeout = NULL;
 	const char *dhgroup = NULL;
+	const char *group_pw = NULL;
 	GString *routes = NULL;
 	GString *uselegacyikeport = NULL;
 	gboolean success = FALSE;
 	guint32 routes_count = 0;
 	gboolean save_password = FALSE;
+	gboolean save_group_password = FALSE;
 	gboolean use_natt = FALSE;
 	gboolean use_force_natt = FALSE;
+	NMSettingSecretFlags flags = NM_SETTING_SECRET_FLAG_NONE;
 
 	s_con = NM_SETTING_CONNECTION (nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION));
 	s_ip4 = (NMSettingIP4Config *) nm_connection_get_setting (connection, NM_TYPE_SETTING_IP4_CONFIG);
@@ -1320,12 +1482,31 @@ export (NMVpnPluginUiInterface *iface,
 		dhgroup = (value[0] == 'd' && value[1] == 'h') ? value + 2 : NULL;
 	}
 
-	value = nm_setting_vpn_get_data_item (s_vpn, NM_VPNC_KEY_XAUTH_PASSWORD_TYPE);
-	if (value && strlen (value)) {
-		if (!strcmp (value, NM_VPNC_PW_TYPE_SAVE))
+	if (nm_setting_get_secret_flags (NM_SETTING (s_vpn), NM_VPNC_KEY_XAUTH_PASSWORD, &flags, NULL)) {
+		if (!(flags & NM_SETTING_SECRET_FLAG_NOT_SAVED))
 			save_password = TRUE;
+	} else {
+		value = nm_setting_vpn_get_data_item (s_vpn, NM_VPNC_KEY_XAUTH_PASSWORD_TYPE);
+		if (value && strlen (value)) {
+			if (!strcmp (value, NM_VPNC_PW_TYPE_SAVE))
+				save_password = TRUE;
+		}
 	}
 
+	/* Group password stuff */
+	if (nm_setting_get_secret_flags (NM_SETTING (s_vpn), NM_VPNC_KEY_SECRET, &flags, NULL)) {
+		if (!(flags & NM_SETTING_SECRET_FLAG_NOT_SAVED))
+			save_group_password = TRUE;
+	} else {
+		value = nm_setting_vpn_get_data_item (s_vpn, NM_VPNC_KEY_SECRET_TYPE);
+		if (value && strlen (value)) {
+			if (!strcmp (value, NM_VPNC_PW_TYPE_SAVE))
+				save_group_password = TRUE;
+		}
+	}
+	if (save_group_password)
+		group_pw = nm_setting_vpn_get_secret (s_vpn, NM_VPNC_KEY_SECRET);
+
 	routes = g_string_new ("X-NM-Routes=");
 	if (s_ip4 && nm_setting_ip4_config_get_num_routes (s_ip4)) {
 		int i;
@@ -1361,7 +1542,7 @@ export (NMVpnPluginUiInterface *iface,
 		 "Host=%s\n"
 		 "AuthType=1\n"
 		 "GroupName=%s\n"
-		 "GroupPwd=\n"
+		 "GroupPwd=%s\n"
 		 "EnableISPConnect=0\n"
 		 "ISPConnectType=0\n"
 		 "ISPConnect=\n"
@@ -1396,10 +1577,12 @@ export (NMVpnPluginUiInterface *iface,
 		 "%s"
 		 "X-NM-Use-NAT-T=%s\n"
 		 "X-NM-Force-NAT-T=%s\n"
+		 "X-NM-SaveGroupPassword=%s\n"
 		 "%s\n",
 		 /* Description */   nm_setting_connection_get_id (s_con),
 		 /* Host */          gateway,
 		 /* GroupName */     groupname,
+		 /* GroupPassword */ group_pw ? group_pw : "",
 		 /* Username */      username != NULL ? username : "",
 		 /* Save Password */ save_password ? "1" : "0",
 		 /* EnableNat */     enablenat ? "1" : "0",
@@ -1410,6 +1593,7 @@ export (NMVpnPluginUiInterface *iface,
 		 /* UseLegacyIKEPort */ (uselegacyikeport->len) ? uselegacyikeport->str : "",
 		 /* X-NM-Use-NAT-T */ use_natt ? "1" : "0",
 		 /* X-NM-Force-NAT-T */ use_force_natt ? "1" : "0",
+		 /* X-NM-SaveGroupPassword */ save_group_password ? "1" : "0",
 		 /* X-NM-Routes */   (routes && routes->str) ? routes->str : "");
 
 	success = TRUE;
@@ -1446,36 +1630,6 @@ get_capabilities (NMVpnPluginUiInterface *iface)
 	return (NM_VPN_PLUGIN_UI_CAPABILITY_IMPORT | NM_VPN_PLUGIN_UI_CAPABILITY_EXPORT);
 }
 
-static gboolean
-delete_connection (NMVpnPluginUiInterface *iface,
-                   NMConnection *connection,
-                   GError **error)
-{
-	NMSettingConnection *s_con;
-	const char *id, *uuid;
-
-	/* Remove any secrets in the keyring associated with this connection's UUID */
-	s_con = (NMSettingConnection *) nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION);
-	if (!s_con) {
-		g_set_error (error,
-		             VPNC_PLUGIN_UI_ERROR,
-		             VPNC_PLUGIN_UI_ERROR_INVALID_CONNECTION,
-		             "missing 'connection' setting");
-		return FALSE;
-	}
-
-	id = nm_setting_connection_get_id (s_con);
-	uuid = nm_setting_connection_get_uuid (s_con);
-
-	if (!keyring_helpers_delete_secret (uuid, VPNC_USER_PASSWORD))
-		g_message ("%s: couldn't delete user password for '%s'", __func__, id);
-
-	if (!keyring_helpers_delete_secret (uuid, VPNC_GROUP_PASSWORD))
-		g_message ("%s: couldn't delete group password for '%s'", __func__, id);
-
-	return TRUE;
-}
-
 static NMVpnPluginUiWidgetInterface *
 ui_factory (NMVpnPluginUiInterface *iface, NMConnection *connection, GError **error)
 {
@@ -1536,7 +1690,6 @@ vpnc_plugin_ui_interface_init (NMVpnPluginUiInterface *iface_class)
 	iface_class->import_from_file = import;
 	iface_class->export_to_file = export;
 	iface_class->get_suggested_name = get_suggested_name;
-	iface_class->delete_connection = delete_connection;
 }
 
 
